-- Copyright 2024 DJ WoodZ

-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:

-- The above copyright notice and this permission notice shall be included in all
-- copies or substantial portions of the Software.

-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
-- SOFTWARE.

EnhancedEconomySettingsGUI = {}

EnhancedEconomySettingsGUI.multiplierOptions = {
  0,
  0.25,
  0.5,
  0.6,
  0.7,
  0.75,
  0.8,
  0.85,
  0.9,
  0.95,
  1,
  1.05,
  1.1,
  1.15,
  1.2,
  1.25,
  1.3,
  1.4,
  1.5,
  1.75,
  2,
  2.5,
  3,
  4,
  5,
  10,
  25,
  50,
  100
}

EnhancedEconomySettingsGUI.frame = nil

local initialised = false

function EnhancedEconomySettingsGUI.init()
  if not initialised then
    -- Multi option input elements
    EnhancedEconomySettingsGUI.multiOptionElements = {}

    -- Hooks into InGameMenuGameSettingsFrame to add/maintain menu items
    InGameMenuGameSettingsFrame.onFrameOpen = Utils.appendedFunction(InGameMenuGameSettingsFrame.onFrameOpen,
      EnhancedEconomySettingsGUI.onFrameOpen)
    InGameMenuGameSettingsFrame.updateGameSettings = Utils.appendedFunction(
      InGameMenuGameSettingsFrame.updateGameSettings, EnhancedEconomySettingsGUI.updateGameSettings)

    initialised = true
  end
end

function EnhancedEconomySettingsGUI.onFrameOpen(frame)
  if EnhancedEconomySettingsGUI.frame == nil then
    EnhancedEconomySettingsGUI:addSubtitleElement(frame, 'EnhancedEconomySettings_Subtitle')

    for _, fillType in pairs(g_fillTypeManager.fillTypes) do
      local value = EnhancedEconomySettings.economySettings[fillType.name]
      local element = EnhancedEconomySettingsGUI:addMultiOptionElement(frame, EnhancedEconomySettingsGUI, fillType.name)

      element.labelElement:setText(EnhancedEconomySettings.economySettingsLabels[fillType.name])

      EnhancedEconomySettingsGUI:setMultiOptionValue(frame, element, value)
    end

    frame.boxLayout:invalidateLayout(false)

    frame.boxLayout.mouseEvent = Utils.overwrittenFunction(frame.boxLayout.mouseEvent,
      EnhancedEconomySettingsGUI.mouseEvent)

    EnhancedEconomySettingsGUI.frame = frame
  elseif (FocusManager:getFocusedElement() ~= nil) then
    -- if FocusManager has a focused element
    local focusedElement = FocusManager:getFocusedElement()

    -- if the focused element is one of our multi options
    local startsWithStr = 'EnhancedEconomySettingsMultiOption_'
    if (focusedElement.id ~= nil
          and focusedElement.id:sub(1, #startsWithStr) == startsWithStr
          and focusedElement.fillTypeName ~= nil
          and EnhancedEconomySettingsGUI.multiOptionElements[focusedElement.fillTypeName]) then
      -- re-request focus to ensure it isn't lost
      FocusManager:unsetFocus(focusedElement)
      FocusManager:setFocus(focusedElement)
    end
  end
end

function EnhancedEconomySettingsGUI.updateGameSettings(frame)
  for elementId, element in pairs(EnhancedEconomySettingsGUI.multiOptionElements) do
    local focusedElement = FocusManager:getFocusedElement()

    local value = EnhancedEconomySettings.economySettings[elementId]
    EnhancedEconomySettingsGUI:setMultiOptionValue(frame, element, value)

    -- if the currently focused element is this element
    local startsWithStr = 'EnhancedEconomySettingsMultiOption_'
    if (focusedElement ~= nil
          and focusedElement.id ~= nil
          and focusedElement.id:sub(1, #startsWithStr) == startsWithStr
          and focusedElement.fillTypeName ~= nil
          and EnhancedEconomySettingsGUI.multiOptionElements[focusedElement.fillTypeName]
          and element.id == focusedElement.id) then
      -- re-request focus to ensure it isn't lost
      FocusManager:unsetFocus(element)
      FocusManager:setFocus(element)
    end
  end
end

function EnhancedEconomySettingsGUI:mouseEvent(super, posX, posY, isDown, isUp, button, ...)
  -- if scroll wheel is being used
  if (button == Input.MOUSE_BUTTON_WHEEL_UP or button == Input.MOUSE_BUTTON_WHEEL_DOWN) then
    -- ensure more accurate scrolling
    if (isUp) then
      local elementCount = #self.elements
      if (elementCount > 0) then
        if (button == Input.MOUSE_BUTTON_WHEEL_UP) then
          local firstVisibleY = self.firstVisibleY - (self.contentSize / elementCount);
          self:scrollTo(math.max(firstVisibleY, 0), true)
        elseif (button == Input.MOUSE_BUTTON_WHEEL_DOWN) then
          local firstVisibleY = self.firstVisibleY + (self.contentSize / elementCount);
          local max = (self.contentSize - self.absSize[2])
          self:scrollTo(math.min(firstVisibleY, max), true)
        end
      end
    end
  else
    return super(self, posX, posY, isDown, isUp, button, ...)
  end
end

function EnhancedEconomySettingsGUI:addMultiOptionElement(frame, target, fillTypeName)
  local element = frame.multiAutoSaveInterval:clone(frame.boxLayout)

  element.id = 'EnhancedEconomySettingsMultiOption_' .. fillTypeName
  element.fillTypeName = fillTypeName
  element.target = target

  element:setCallback("onClickCallback", "handleMultiOptionUpdate")
  element:setVisible(true)

  -- ensures element can be keyboard focused
  element.focusId = FocusManager.serveAutoFocusId()
  FocusManager:loadElementFromCustomValues(element, element.focusId)
  for index, subElement in pairs(element.elements) do
    subElement.id = 'EnhancedEconomySettingsMultiOption_' .. fillTypeName .. '_' .. index
    subElement.focusId = FocusManager.serveAutoFocusId()
    FocusManager:loadElementFromCustomValues(subElement, subElement.focusId)
  end

  element.labelElement:setText(fillTypeName)

  element.texts = {}
  for index, value in pairs(EnhancedEconomySettingsGUI.multiplierOptions) do
    element.texts[index] = EnhancedEconomySettingsGUI:getText('EnhancedEconomySettings_Multi' ..
      EnhancedEconomySettings:round(value, 2))
    if (value == 1) then
      element:setState(index)
    end
  end

  local hudOverlayFilename = g_fillTypeManager.nameToFillType[fillTypeName].hudOverlayFilename
  if (hudOverlayFilename ~= nil and hudOverlayFilename ~= '') then
    local iconElement = BitmapElement.new(target)
    iconElement:applyProfile('ingameMenuPriceGoodsIcon')
    iconElement.id = fillTypeName
    iconElement:setImageFilename(hudOverlayFilename)

    element.addElement(element, iconElement)

    element.labelElement.textOffset[1] = 0.025
    element.labelElement.textFocusedOffset[1] = 0.025
    element.labelElement.textMaxWidth = element.labelElement.textMaxWidth - element.labelElement.textOffset[1]
  end

  EnhancedEconomySettingsGUI.multiOptionElements[fillTypeName] = element

  frame.boxLayout:addElement(element)

  return element
end

function EnhancedEconomySettingsGUI:setMultiOptionValue(frame, element, value)
  if (type(value) == 'number' and EnhancedEconomySettings:tableContains(EnhancedEconomySettingsGUI.multiplierOptions, value)) then
    if frame.hasMasterRights then
      element.elements[6]:setText(EnhancedEconomySettingsGUI:getText('EnhancedEconomySettings_Tooltip'))
      element:setDisabled(false)
    else
      element.elements[6]:setText(EnhancedEconomySettingsGUI:getText('EnhancedEconomySettings_TooltipNoPermission'))
      element:setDisabled(true)
    end
    element:setState(EnhancedEconomySettings:tableContains(EnhancedEconomySettingsGUI.multiplierOptions, value))
  else
    element.elements[6]:setText(EnhancedEconomySettingsGUI:getText('EnhancedEconomySettings_TooltipLocked'))
    -- Set to 0 as free anyway
    element:setState(EnhancedEconomySettings:tableContains(EnhancedEconomySettingsGUI.multiplierOptions, 0))
    element:setDisabled(true)
  end
end

function EnhancedEconomySettingsGUI:addSubtitleElement(frame, text)
  local element = TextElement.new()
  element:applyProfile('settingsMenuSubtitle', true)
  element:setText(EnhancedEconomySettingsGUI:getText(text))
  frame.boxLayout:addElement(element)
end

function EnhancedEconomySettingsGUI:getText(key)
  local result = g_i18n.modEnvironments[EnhancedEconomySettings.name].texts[key]
  if result == nil then
    return g_i18n:getText(key)
  end
  return result
end

function EnhancedEconomySettingsGUI:handleMultiOptionUpdate(state, element)
  EnhancedEconomySettings.economySettings[element.fillTypeName] = EnhancedEconomySettingsGUI.multiplierOptions[state]
  EnhancedEconomySettings:updateEconomy(EnhancedEconomySettings.economySettings)
  EnhancedEconomySettingsUpdateEvent:sendEvent(EnhancedEconomySettings.economySettings)

  -- re-request focus to ensure it isn't lost
  FocusManager:unsetFocus(element)
  FocusManager:setFocus(element)
end
